/*
 * Copyright 2006, Thomas Maier-Komor
 * LICENSE: GPLv2, see file COPYING for details
 */

#include "fsstats.h"
#include "log.h"

#include <alloca.h>
#include <assert.h>
#include <fcntl.h>
#include <limits.h>
#include <stdio.h>
#include <stdlib.h>
#include <strings.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <sys/statvfs.h>

unsigned long long FsFree = 0, FsSize = 0;

static fsstat_t *Fs = 0;


static void
add_fs(const char *dev, const char *mnt, const char *typ,
		unsigned long long blocks, unsigned long long free)
{
	fsstat_t *f = Fs;
	
	while (f && strcmp(dev,f->dev))
		f = f->next;
	if (f) {
		f->blocks = blocks;
		f->free = free;
	} else {
		f = malloc(sizeof(fsstat_t));
		f->next = Fs;
		f->dev = strdup(dev);
		f->mnt = strdup(mnt);
		strcpy(f->typ,typ);
		f->blocks = blocks;
		f->free = free;
		Fs = f;
	}
}


static int
Crit(fsstat_t *a, fsstat_t *b)
{
	return a->free > b->free;
}


static void
sort_fs(void)
{
	fsstat_t *f = Fs, *h = 0, *t = 0;
	while (f) {
		fsstat_t *n = f->next, *i = h;
		if (f->blocks == 0) {
			free(f->dev);
			free(f->mnt);
			free(f);
			f = n;
			continue;
		}
		while (i && Crit(f,i)
			&& i->next && Crit(f,i->next)) {
			assert(i != i->next);
			i = i->next;
		}
		if (i == 0) {
			if (t) {
				assert(h && !Crit(t, i));
				t->next = f;
				t = f;
			} else {
				assert(h == 0);
				h = t = f;
			}
			f->next = 0;
		} else {
			if (Crit(f,i)) {
				f->next = i->next;
				i->next = f;
			} else {
				f->next = i;
				h = f;
			}
			assert(i != i->next);
		}
		f = n;
		assert(f == 0 || f != f->next);
	}
	Fs = h;
}


void
update_fs_stats(void)
{
	static int upd = 0;
	FILE *mnttab;
	struct statvfs vst;
	fsstat_t *f = Fs;
	char device[PATH_MAX], mnt_pnt[PATH_MAX],
	       	fstyp[16];	/* Is sizeof fstyp OK? */

	if (upd) {
		--upd;
		return;
	}
	FsFree = 0;
	FsSize = 0;
	while (f) {
		f->blocks = 0;	/* marks fs that have been umounted for
				   removal */
		f = f->next;
	}
	mnttab = fopen("/etc/mnttab","r");
	assert(mnttab != 0);
	while (3 == fscanf(mnttab,"%[^\t]\t%[^\t]\t%[^\t]\t%*[^\n]\n",
				device, mnt_pnt, fstyp)) {
		if ((0 == strcmp("ufs",fstyp)) || (0 == strcmp("pcfs",fstyp))) {
			if (0 == statvfs(mnt_pnt,&vst)) {
				/*
				dbug("%s @ %s : %llu/%llu\n",
					fstyp,
					mnt_pnt,
					(unsigned long long)vst.f_bfree,
					(unsigned long long)vst.f_blocks);
				*/
				FsFree += ((unsigned long long)vst.f_bavail * vst.f_frsize) >> 10;
				FsSize += ((unsigned long long)vst.f_blocks * vst.f_frsize) >> 10;
				add_fs(device,mnt_pnt,fstyp,vst.f_blocks,vst.f_bavail);
			}
		}
	}
	fclose(mnttab);
	sort_fs();
	upd = 10;
}


fsstat_t *
get_fs_stats(void)
{
	return Fs;
}

