/*
 * Copyright 2006, Thomas Maier-Komor
 * LICENSE: GPLv2, see file COPYING for details
 */

#include <assert.h>
#include <errno.h>
#include <fcntl.h>
#include <poll.h>
#include <signal.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <sys/stat.h>
#include <sys/wait.h>
#include <unistd.h>

#include "log.h"
#include "network.h"
#include "sysstat.h"
#include "terminal.h"


int Fslines = 5;
unsigned Interval = 5;

static sysstat_t
	*SS = 0, *Current = 0;
static unsigned
	Curses = 1;

static void
dealloc(sysstat_t *s)
{
	userstat_t *us;
	fsstat_t *fs;
	if (s == 0)
		return;
	us = s->us;
	while (us) {
		userstat_t *n = us->next;
		free(us->username);
		free(us);
		us = n;
	}
	fs = s->fs;
	while (fs) {
		fsstat_t *n = fs->next;
		free(fs->dev);
		free(fs->mnt);
		free(fs);
		fs = n;
	}
	free(s);
}


static void
add_stats(sysstat_t *s)
{
	sysstat_t *i = SS;
	assert((s != i) && (s->next == 0) && (s->prev == 0));
	/* search, if we already have an entry for this host */
	while (i && s) {
		assert(i->next && i->prev);
		if (0 == strcmp(i->Hostinfo,s->Hostinfo)) {
			/* replace this entry */
			if (i->prev != i) {
				assert(i->next != i);
				s->prev = i->prev;
				assert(s->prev->next == i);
				s->prev->next = s;
				s->next = i->next;
				assert(s->next->prev == i);
				s->next->prev = s;
			} else {
				assert(i->next == i);
				s->prev = s;
				s->next = s;
			}
			if (Current == i)
				Current = s;
			if (SS == i)
				SS = s;
			dealloc(i);
			assert(s->prev && s->next);
			return;
		}
		i = i->next;
		if (i == SS)
			break;
	}
	/* new sysstat not already in our list */
	assert(s);
	if (SS) {
		s->next = SS;
		s->prev = SS->prev;
		assert(s->prev->next == SS);
		s->prev->next = s;
		SS->prev = s;
	} else {
		s->next = s;
		s->prev = s;
		SS = s;
		Current = s;
	}
	assert(s->prev && s->next);
}


static void
process_input(int s, int itv)
{
	struct pollfd fds[2];
	char c;
	sysstat_t *st;
	
	fds[0].fd = STDIN_FILENO;
	fds[0].events = POLLIN;
	fds[1].fd = s;
	fds[1].events = POLLIN;
	if (-1 == poll(fds,s == -1 ? 1 : sizeof(fds)/sizeof(struct pollfd),itv)) {
		if (errno != EINTR)
			error("error polling: %s\n",strerror(errno));
		return;
	}
	if (1 == read(STDIN_FILENO,&c,sizeof(char))) {
		switch (c) {
		case 'q':
			exit(EXIT_SUCCESS);
			break;
		case 'h':
			if (Current == 0)
				Current = SS;
			else
				Current = Current->next;
			break;
		case 'H':
			if (Current == 0)
				Current = SS;
			else
				Current = Current->prev;
			break;
		case 'n':
			set_sort_by_nproc();
			break;
		case 'r':
			set_sort_by_rss();
			break;
		case 's':
			set_sort_by_size();
			break;
		case 't':
			set_sort_by_time();
			break;
		case 'u':
			set_sort_by_user();
			break;
		case 'c':
			set_sort_by_pcpu();
			break;
		case '?':
			if (Curses)
				display_help();
		default:
			;
		}
	}
	if ((s != -1) && (st = receive_stats(s))) {
		assert(st != SS);
		add_stats(st);
	}
}


static void
usage(int r)
{
	printf(
	"sysstat [option] [interval [count]]\n"
	"Options:\n"
	"-h        : display this help screen and exit\n"
	"-c        : write stats to stdout instead of using curses\n"
	"-d        : start sysstat as daemon, multicasting stats\n"
	"-f <num>  : show at least num lines of filesystem information\n"
	"-l        : start sysstat in local-mode (i.e. without network support)\n"
	"-m <addr> : send info to multicast address <addr> (default: 224.0.0.75)\n"
	"-p <port> : set multicast port to <port> (default: 7536)\n"
	"-r	   : sort user/process summary by rss\n"
	"-s      : sort user/process summary by size\n"
	"-t      : sort user/process summary by %%CPU time\n"
	"-u      : sort user/process summary by username\n"
	"-v      : operate in verbose mode\n"
	"-V      : display version number and exit\n"
	);
	exit(r);
}


static void
version(void)
{
	printf(
	"sysstat version "VERSION"\n"
	"Copyright 2006, Thomas Maier-Komor\n"
	"This software is licensed under the GPLv2. See file COPYING for details.\n"
	);
	exit(EXIT_SUCCESS);
}


int
main(int argc, char **argv)
{
	unsigned daemon = 0, multicast = 0, localmode = 0,
		 iter = 0, killdaemon = 0;
	int c, count = -1, pid, sock;
	extern char *optarg;
	FILE *pidf;

	while ((c = getopt(argc, argv, "cdf:hklm:p:rstuvV")) != -1) {
		extern int Port;
		extern char *McastAddr;

		switch(c) {
		default:
			assert(0);
			break;
		case 'c':
			Curses = 0;
			break;
		case 'd':
			daemon = 1;
			multicast = 1;
			break;
		case 'f':
			if (1 != sscanf(optarg,"%u",&Fslines))
				usage(EXIT_FAILURE);
			break;
		case 'h':
			usage(EXIT_SUCCESS);
			break;
		case 'k':
			killdaemon = 1;
			break;
		case 'l':
			localmode = 1;
			break;
		case 'm':
			McastAddr = optarg;
			info("multicast address set to %s\n",McastAddr);
			break;
		case 'p':
			if (1 != sscanf(optarg,"%u",&Port))
				usage(EXIT_FAILURE);
			break;
		case 'r':
			set_sort_by_rss();
			break;
		case 's':
			set_sort_by_size();
			break;
		case 't':
			set_sort_by_time();
			break;
		case 'u':
			set_sort_by_user();
			break;
		case 'v':
			Verbose = 1;
			break;
		case 'V':
			version();
			break;
		case '?':
			usage(EXIT_FAILURE);
			break;
		}
	}
	if (optind < argc) {
		if (1 != sscanf(argv[optind],"%u",&Interval))
			usage(EXIT_FAILURE);
		if ((Interval != 0) && (daemon == 0) && (localmode == 0))
			dbug("interval set to %u\n",Interval);
		++optind;
	}
	if (optind < argc) {
		if (1 != sscanf(argv[optind],"%d",&count) || (count <= 0))
			usage(EXIT_FAILURE);
		++optind;
	}
	if (optind < argc) {
		usage(EXIT_FAILURE);
	}
	if (strcmp(argv[0], "sysstatd") == 0) {
		multicast = 1;
		daemon = 1;
	}
	if (localmode) {
		sysstat_t SS;
		SS.next = &SS;
		SS.prev = &SS;
		init_localstat(&SS);
		if (Curses)
			init_curses();
		for (;count;--count) {
			get_local_data(&SS);
			if (Curses) {
				display_curses(&SS);
			} else {
				display_stdout(&SS);
			}
			process_input(-1,Interval);
		}
	}
	pidf = fopen("/tmp/sysstat.pid","r");
	if (pidf) {
		char pn[] = "/proc/\0\0\0\0\0\0",*p;
		struct stat st;
		info("/tmp/sysstat.pid exists\n");
		if (1 != fscanf(pidf,"%d",&pid)) {
			warn("corrupt pid file /tmp/systat.pid - unlinking it\n");
			if (-1 == unlink("/tmp/sysstat.pid"))
				warn("cannot unlink stale pid file /tmp/sysstat.pid: %s\n",strerror(errno));
			multicast = 1;
		} else {
			p = pn + 6;
			sprintf(p,"%d",pid);
			info("stating %s\n",pn);
			if (-1 == stat(pn,&st)) {
				info("unlinking stale pid file /tmp/sysstat.pid\n");
				if (-1 == unlink("/tmp/sysstat.pid"))
					warn("cannot unlink stale pid file /tmp/sysstat.pid: %s\n",strerror(errno));
				multicast = 1;
			} else {
				info("daemon already running with pid %d\n",pid);
				if (daemon)
					exit(EXIT_FAILURE);
				if (killdaemon) {
					if (0 == kill(pid,SIGTERM)) {
						info("sending SIGTERM to %d\n",pid);
						exit(EXIT_SUCCESS);
					} else {
						error("cannot send SIGTERM to %d: %s\n",pid,strerror(errno));
						exit(EXIT_FAILURE);
					}
				}
			}
		}
		fclose(pidf);
	} else {
		multicast = 1;
	}
	if (killdaemon) {
		error("cannot kill sysstat daemon: no sysstat daemon running\n");
		exit(EXIT_FAILURE);
	}
        if (multicast) {
		pid_t pid;
		if ((pid = fork())) {
			int status;
			waitpid(pid, &status, 0);
			assert(status == 0);
		} else {
			pid_t dpid;
			if (dpid = fork()) {
				FILE *pidf;
				pidf = fopen("/tmp/sysstat.pid","w");
				if (pidf == 0)
					error("cannot create pid file: %s\n",strerror(errno));
				fprintf(pidf,"%d\n",dpid);
				fclose(pidf);
				chmod("/tmp/sysstat.pid",0666);
				exit(EXIT_SUCCESS);
			} else
				rundaemon(Interval);
		} 
		if (daemon)
			exit(EXIT_SUCCESS);
	} else
		printf("sysstat daemon already running: waiting for data...\n");

	sock = setup_socket();
	if (-1 == fcntl(STDIN_FILENO,F_SETFL, O_NONBLOCK))
		error("Cannot set input file to non-blocking: %s\n",strerror(errno));
	if (Curses)
		Curses = init_curses();
	while ((count == 0) || (iter < count)) {
		if (Curses) {
			iter += display_curses(Current);
		} else {
			iter += display_stdout(Current);
		}
		process_input(sock,-1);
	}
	/*return 0; 	-- lint wants it, cc doesn't... */
}

// vim:tw=0
