/*
 * Copyright 2006, Thomas Maier-Komor
 * LICENSE: GPLv2, see file COPYING for details
 */

#include "log.h"
#include "sysstat.h"

#include <alloca.h>
#include <assert.h>
#include <ncurses.h>
#include <errno.h>
#include <math.h>
#include <signal.h>
#include <stdlib.h>
#include <string.h>
#include <term.h>
#include <unistd.h>


static WINDOW *w = 0, *H = 0;
enum { BLACK_ON_WHITE = 1, BLUE_ON_WHITE, RED_ON_WHITE };
static int Sigwinch = 1, stress, nostress, header, underline, 
    Numlines = 0, Numcol = 0;
static sysstat_t *displayed = 0;
static clock_t Dis_ts;
extern int Fslines;

void
exit_curses(void)
{
	endwin();
}


void
sighandler(int sig)
{
	if (sig == SIGINT || sig == SIGTERM) {
		exit(EXIT_SUCCESS);
	}
	assert(sig == SIGWINCH);
	Sigwinch = 1;
}


void
display_help(void)
{
	char *line;

	if (H == 0) {
		if ((Numlines == 0) || (Numcol == 0)) {
			struct winsize w;
			if (-1 == ioctl(STDOUT_FILENO,TIOCGWINSZ,&w))
				error("cannot read terminal size: %s\n",strerror(errno));
			Numlines = w.ws_row;
			Numcol = w.ws_col;
		}
		assert(Numlines && Numcol);
		H = newwin(Numlines,Numcol,0,0);
		line = alloca(Numcol + 1);
		wclear(H);
		memset(line,'=',Numcol);
		line[Numcol] = 0;
		mvwaddstr(H, 0,0,"sysstat version "VERSION);
		mvwaddstr(H, 1,0,"Copyright 2006, Thomas Maier-Komor");
		mvwaddstr(H, 2,0,"license: GPLv2 - see file COPYING or \"man gpl\" for details");
		mvwaddstr(H, 3,0,line);

		mvwaddstr(H, 5,0,"Hotkeys:");
		mvwaddstr(H, 6,0,"--------");

		mvwaddstr(H, 7,0,"'q' : quit sysstat");
		mvwaddstr(H, 8,0,"'?' : toggle this help screen");

		mvwaddstr(H,10,0,"'h' : display next host");
		mvwaddstr(H,11,0,"'H' : display previous host");

		mvwaddstr(H,13,0,"'u' : sort process list by username");
		mvwaddstr(H,14,0,"'n' : sort process list by NPROC");
		mvwaddstr(H,15,0,"'r' : sort process list by RSS");
		mvwaddstr(H,16,0,"'s' : sort process list by SIZE");
		mvwaddstr(H,17,0,"'t' : sort process list by Time");
		mvwaddstr(H,18,0,"'c' : sort process list by %CPU");
		wsyncup(H);
		wrefresh(H);
	} else {
		delwin(H);
		H = 0;
	}
}


int
init_curses(void)
{
	void (*osig)(int);

	if (0 == isatty(STDOUT_FILENO))
		return 0;
	w = initscr();
	noecho();
	cbreak();
	nodelay(w, TRUE);
	nonl();
	curs_set(0);
	if (has_colors()) {
		start_color();
		init_pair(BLACK_ON_WHITE,COLOR_BLACK,COLOR_WHITE);
		init_pair(BLUE_ON_WHITE,COLOR_BLUE,COLOR_WHITE);
		init_pair(RED_ON_WHITE,COLOR_RED,COLOR_WHITE);
		nostress = COLOR_PAIR(BLACK_ON_WHITE);
		stress = A_BOLD | COLOR_PAIR(RED_ON_WHITE);
		header = COLOR_PAIR(BLUE_ON_WHITE) | A_UNDERLINE;
		underline = COLOR_PAIR(BLACK_ON_WHITE) | A_UNDERLINE;
	} else {
		header = A_UNDERLINE;
		underline = A_UNDERLINE;
		nostress = 0;
		stress = A_BOLD;
	}
	osig = signal(SIGWINCH,sighandler);
	assert(osig != SIG_ERR);
	osig = signal(SIGINT,sighandler);
	assert(osig != SIG_ERR);
	osig = signal(SIGTERM,sighandler);
	assert(osig != SIG_ERR);
	atexit(exit_curses);
	return 1;
}


static char *
per2str(double d)
{
	static char str[7];
	assert(d <= 1E2);
	if (d < 0) {
		memcpy(str,"n/a ",5);
	} else if (d < 2E-2) {
		sprintf(str,"%4.0f",d);
	} else if (d < 9.995) {
		sprintf(str,"%4.2f",d);
	} else if (d < 99.95) {
		sprintf(str,"%4.1f",d);
	} else
		sprintf(str,"%4.0f",rint(d));
	return str;
}


static char *
dbl2str(double d)
{
	static char str[7];
	double r = rint(d);
	if (d < 0) {
		memcpy(str,"unav.",6);
	} else if (d < 2E-2) {
		sprintf(str,"%4.0f ",d);
	} else if (d < 9.995) {
		sprintf(str,"%4.2f ",d);
	} else if (r < 1E2) {
		sprintf(str,"%4.1f ",d);
	} else if (r < 1E4) {
		sprintf(str,"%4.0f ",r);
	} else if (d < 10E6) {
		d /= 1E3;
		sprintf(str,"%4.0fk",rint(d));
	} else {
		d /= 1E6;
		sprintf(str,"%4.0fM",rint(d));
	}
	return str;
}


static char *
kb2str(long long ll)
{
	static char str[7];
	char dim = ' ';
	double v;

	if (ll < 0) {
		memcpy(str,"???",4);
		return str;
	} else if (ll >= 10000LLU*1024LLU*1024LLU) {
		v = (double)(ll >> 20)/1024;
		dim = 'T';
	} else if (ll >= 10000*1024) {
		v = (double)(ll >> 10)/1024;
		dim = 'G';
	} else if (ll >= 10000) {
		v = (double)ll/1024;
		dim = 'M';
	} else {
		v = (double) ll;
		dim = 'k';
	}
	if (rint(v) >= 100)
		sprintf(str,"%4.0f%c",v,dim);
	else
		sprintf(str,"%4.1f%c",v,dim);
	return str;
}


unsigned 
display_curses(sysstat_t *s)
{
	int num;
	userstat_t *u;
	fsstat_t *fs;
	char line[4096], *l;
	int fslines = Fslines, ulines = 0, unshown = 0;
	sysstat_t *h = s;

	if (H || (s == 0) ||
	    ((s == displayed) && (s->TS == Dis_ts) && (Sigwinch == 0)))
		return 0;
	displayed = s;
	Dis_ts = s->TS;
	h = s->next;
	while (h != s) {
		++unshown;
		h = h->next;
	}
	assert(w);
	if (Sigwinch) {
		struct winsize w;
		if (-1 == ioctl(STDOUT_FILENO,TIOCGWINSZ,&w))
			error("cannot read terminal size: %s\n",strerror(errno));
		Numlines = w.ws_row;
		Numcol = w.ws_col;
		dbug("SIGWINCH: Numlines = %d\n",Numlines);
		endwin();
		refresh();
		Sigwinch = 0;
		(void) signal(SIGWINCH, sighandler);
	}
	if (unshown) 
		--Numlines;
	num = 0;
	if (fslines > s->numfs) {
		dbug("fslines = %d, numfs = %d\n",fslines,s->numfs);
		fslines = s->numfs;
	}
	attrset(nostress);
	clear();
	attrset(header);
	memset(line,' ',80);
	/*sprintf(line,"12345678901234567890123456789012345678901234567890123456789012345678901234567890\n");*/
	strcpy(line,s->Hostinfo);
	l = line + strlen(s->Hostinfo);
	if (s->Ocpus == s->Ncpus) {
		if (s->Ncpus > 1) {
			l += sprintf(l,", %d CPUs", s->Ocpus);
		} else {
			assert(s->Ocpus == 1);
			l += sprintf(l,", 1 CPU");
		}
	} else {
		assert(s->Ncpus > 1);
		l += sprintf(l,", %d/%d CPUs",s->Ocpus,s->Ncpus);
	}
	addstr(line);
	if (s->Devices != 1) {
		l += sprintf(l,", %lu disks ", s->Devices);
	} else {
		l += sprintf(l,", 1 disk ");
	}
	*l = ' ';
	strftime(line+61,20,"%c",localtime(&s->TS));
	mvaddstr(num++,0,line);
	if (num == Numlines) goto done;

	attrset(nostress);
	/*-------------- CPU -------------*/
	sprintf(line," User   CPU: %s%% | ", per2str(s->UserP));
	mvaddstr(num++,0,line);
	sprintf(line,"LoadAv/1m:   %4.2f | ", s->loadavg[0]);
	addstr(line);
	sprintf(line,"intr/s:     %s | ", dbl2str(s->IntrR));
	addstr(line);
	sprintf(line,"pswitch/s:   %s ", dbl2str(s->PswitchR));
	addstr(line);
	if (num == Numlines) goto done;

	sprintf(line," Kernel CPU: %s%% | ",per2str(s->KernelP));
	mvaddstr(num++,0,line);
	sprintf(line,"LoadAv/5m:   %4.2f | ", s->loadavg[1]);
	addstr(line);
	sprintf(line,"intrblk/s:  %s | ", dbl2str(s->IntrBlkR));
	addstr(line);
	sprintf(line,"migrate/s:   %s ", dbl2str(s->MigrateR));
	addstr(line);
	if (num == Numlines) goto done;

	if (s->IdleP <= 1)
		attrset(stress | A_UNDERLINE);
	attrset(underline);
	mvaddstr(num++,0," Idle   CPU: ");
	if (s->IdleP <= 1)
		attrset(stress | A_UNDERLINE);
	sprintf(line,"%s%%",per2str(s->IdleP));
	addstr(line);
	attrset(underline);
	sprintf(line," | LoadAv/15m:  %4.2f | ", s->loadavg[2]);
	addstr(line);
	sprintf(line,"xcalls/s:   %s | ", dbl2str(s->XcallsR));
	addstr(line);
	sprintf(line,"inv_swtch/s: %s ", dbl2str(s->InvSwtchR));
	addstr(line);
	if (num == Numlines) goto done;

	/*-------------- Mem -------------*/
	attrset(nostress);
	sprintf(line," Physical:   %s | ", kb2str(s->Physmem));
	mvaddstr(num++,0,line);
	sprintf(line,"SwapTotal:  %s | ", kb2str(s->SwapTotal));
	addstr(line);
	sprintf(line,"Locked:     %s | ", kb2str(s->Pageslocked));
	addstr(line);
	sprintf(line,"pgswapin/s:  %s ", dbl2str(s->PgswapinR));
	addstr(line);
	if (num == Numlines) goto done;

	attrset(underline);
	sprintf(line," Free:       %s | ", kb2str(s->Pagesfree));
	mvaddstr(num++,0,line);
	sprintf(line,"SwapFree:   %s | ", kb2str(s->SwapFree));
	addstr(line);
	addstr("scan/s:     ");
	if (s->ScanR)
		attrset(A_UNDERLINE|stress);
	addstr(dbl2str(s->ScanR));
	if (s->ScanR)
		attrset(underline);
	sprintf(line," | pgswapout/s: %s ", dbl2str(s->PgswapoutR));
	addstr(line);
	if (num == Numlines) goto done;

	/*-------------- I/O -------------*/
	attrset(nostress);
	sprintf(line," syscall/s:  %s | ", dbl2str(s->SyscallR));
	mvaddstr(num++,0,line);
	sprintf(line,"sysfork/s:  %s | ", dbl2str(s->SysforkR));
	addstr(line);
	sprintf(line,"sysread/s:  %s | ", dbl2str(s->SysreadR));
	addstr(line);
	sprintf(line,"IP rcv/s:    %s ", dbl2str(s->IpInR));
	addstr(line);
	if (num == Numlines) goto done;

	attrset(underline);
	sprintf(line," nfsd ops/s: %s | ", dbl2str(s->NfsdR));
	mvaddstr(num++,0,line);
	sprintf(line,"sysexec/s:  %s | ", dbl2str(s->SysexecR));
	addstr(line);
	sprintf(line,"syswrite/s: %s | ", dbl2str(s->SyswriteR));
	addstr(line);
	sprintf(line,"IP send/s:   %s ", dbl2str(s->IpOutR));
	addstr(line);
	if (num == Numlines) goto done;

	attrset(nostress);
	sprintf(line," Storage:    %s | ", kb2str(s->Storage));
	mvaddstr(num++,0,line);
	sprintf(line,"net errs: %6lu  | ",
			s->IpInDiscards+s->IpOutDiscards+s->IpInErrs);
	addstr(line);
	sprintf(line,"DiskIn/s:   %s | ", kb2str((long long)s->DiskIR));
	addstr(line);
	sprintf(line,"NfsIn/s:     %s ", kb2str((long long)s->NfsIR));
	addstr(line);
	if (num == Numlines) goto done;

	attrset(underline);
	sprintf(line," FsFree:     %s | ", kb2str(s->FsFree));
	mvaddstr(num++,0,line);
	sprintf(line,"disk errs: %5lu  | ", s->HardErrors+s->SoftErrors);
	addstr(line);
	sprintf(line,"DiskOut/s:  %s | ", kb2str((long long)s->DiskOR));
	addstr(line);
	sprintf(line,"NfsOut/s:    %s ", kb2str((long long)s->NfsOR));
	addstr(line);
	if (num == Numlines) goto done;

	ulines = Numlines - num - fslines;
	if (ulines < fslines) {
		fslines = (Numlines - num) / 2;
		ulines = Numlines - num - fslines;
		assert(ulines >= fslines);
	}
	dbug("fslines = %d\n",fslines);
	attrset(nostress);
	if ((num + 1) < (Numlines - fslines))
		mvaddstr(num++,0,"User           NPROC ( O/ S/ R/ Z/ T) NLWP   SIZE    RSS  %MEM  %CPU    CPU-TIME");
	s->us = sort_users(s->us);
	u = s->us;
	while (u && (num < (Numlines-fslines))) {
		long h,m,s;
		unsigned ul = 0;
		if ((u->next == 0) || (num == (Numlines - fslines - 1)))
			attrset(nostress | (ul = A_UNDERLINE));
		assert(u->time.tv_sec >= 0);
		assert(u->time.tv_nsec >= 0);
		s = u->time.tv_sec + rint((double)u->time.tv_nsec/1000000000);
		m = u->time.tv_sec / 60;
		s %= 60;
		h = m / 60;
		m -= h * 60;
		assert((s >= 0) && (s < 60) && (m >= 0) && (m < 60) && (h >= 0));
		sprintf(line,"%-15s %4d (%2d/%2d/",
				u->username,
				u->nproc,
				u->oncpu,
				u->sleep);
		mvaddstr(num,0,line);
		if (u->runnable)
			attrset(stress | ul);
		sprintf(line,"%2d", u->runnable);
		addstr(line);
		attrset(nostress | ul);
		sprintf(line,"/%2d/%2d) %4d  %s  ",
				u->zombie,
				u->stopped,
				u->nlwp,
				kb2str(u->size));
		addstr(line);
		sprintf(line,"%s  %3.f%%  %3.f%%  %4d:%02d:%02d",
				kb2str(u->rss),
				u->pctmem,
				u->pctcpu,
				h,m,s);
		addstr(line);
		u = u->next;
		dbug("us@%d\n",num);
		++num;
	}
	fs = s->fs;
	attrset(nostress);
	while ((num < Numlines) && fs) {
		char *l = line;
		if (unshown) {
			if ((num + 1 == Numlines) || (0 == fs->next))
				attrset(underline);
		}
		if (strlen(fs->dev) > 30)
			l += sprintf(l,"%.27s...  %-4s ",fs->dev,fs->typ);
		else
			l += sprintf(l,"%-30s  %-4s ",fs->dev,fs->typ);
		if (strlen(fs->mnt) > 30)
			l += sprintf(l,"%.27s...  %s",fs->mnt,kb2str(fs->free));
		else
			l += sprintf(l,"%-30s  %s/",fs->mnt,kb2str(fs->free));
		strcpy(l,kb2str(fs->blocks));
		mvaddstr(num,0,line);
		fs = fs->next;
		dbug("fs@%d:%s\n",num,line);
		++num;
	}
done:
	if (unshown) {
		sysstat_t *u = s->next;
		char *L = line;
		size_t maxlen = sizeof(line) - 1;
		
		attrset(nostress);
		if (Numcol < maxlen)
			maxlen = Numcol;
		++Numlines;
		strcpy(L,"hosts: ");
		L += 7;
		while (u != s) {
			char *d,*b;
			size_t l;

			b = strchr(u->Hostinfo,' ');
			d = strchr(u->Hostinfo,'.');
			if (b && d) {
				if (b < d)
					l = b - u->Hostinfo;
				else
					l = d - u->Hostinfo;
			} else if (b) {
				l = b - u->Hostinfo;
			} else if (d) {
				l = d - u->Hostinfo;
			} else {
				assert(0);
			}
			if (L + l - line >= maxlen)
				break;
			strncpy(L,u->Hostinfo,l);
			L += l;
			u = u->next;
			if (u != s) {
				if (L + 2 - line >= maxlen)
					break;
				strcpy(L,", ");
				L += 2;
			}
		}
		*L = 0;
		mvaddstr(num,0,line);
	}
	move(0,0);
	refresh();
	return 1;
}
