/*
 * Copyright 2006, Thomas Maier-Komor
 * LICENSE: GPLv2, see file COPYING for details
 */

#include "log.h"
#include "userstat.h"

#include <assert.h>
#include <dirent.h>
#include <errno.h>
#include <fcntl.h>
#include <limits.h>
#include <procfs.h>
#include <pwd.h>
#include <stdlib.h>
#include <strings.h>
#include <sys/mman.h>
#include <sys/stat.h>
#include <unistd.h>


static userstat_t *UserHead = 0;
size_t Nproc, Nlwp;
static int sort_nproc(userstat_t *u1, userstat_t *u2);
static int (*Crit)(userstat_t *, userstat_t *) = sort_nproc;


static userstat_t *
get_userstat(uid_t uid)
{
	userstat_t *u = UserHead;
	while ((u != 0) && (u->uid != uid))
		u = u->next;
	if (u == 0) {
		struct passwd *p;
		u = calloc(1,sizeof(userstat_t));
		assert(u);
		u->uid = uid;
		if ((p = getpwuid(uid)))
			u->username = strdup(p->pw_name);
		else
			u->username = strdup("<unknown>");
		u->next = UserHead;
		UserHead = u;
	}
	return u;
}


static void 
update_procinfo(void)
{
	static DIR *procdir = 0;
	struct dirent *ent;
	userstat_t *u = UserHead;
	if (0 == procdir) {
		procdir = opendir("/proc");
		assert(procdir);
	}
	while (u) {
		u->nproc = 0;
		u->nlwp = 0;
		u->oncpu = 0;
		u->sleep = 0;
		u->zombie = 0;
		u->runnable = 0;
		u->stopped = 0;
		u->rss = 0;
		u->size = 0;
		u->pctcpu = 0;
		u->pctmem = 0;
		u->time.tv_sec = 0;
		u->time.tv_nsec = 0;
		u = u->next;
	}
	u = UserHead;
	Nproc = 0;
	Nlwp = 0;
	rewinddir(procdir);
	while (ent = readdir(procdir)) {
		char psinfo[PATH_MAX] = "/proc/";
		int psi,n;
		psinfo_t p;
		userstat_t *u;
		long long ns;

		switch(ent->d_name[0]) {
		case '.':
			switch (ent->d_name[1]) {
			case '.': if (ent->d_name[2] != '\0')
					  break;
				  /*FALLTHROUGH*/
			case '\0':
				continue;
			default:;
			}
			/*FALLTHROUGH*/
		default:;
		}
		/* I think, the above is a little bit more elegant than
		 * this... ;-)
		if (0 == strcmp(".",ent->d_name))
			continue;
		if (0 == strcmp("..",ent->d_name))
			continue;
		*/
		strcpy(psinfo+6,ent->d_name);
		strcat(psinfo+7,"/psinfo");
		psi = open(psinfo,O_RDONLY);
		if (-1 == psi) {
			/* process died after reading directory entry */
			continue;
		}
		n = read(psi,&p,sizeof(p));
		assert(n == sizeof(p));
		++Nproc;
		Nlwp += p.pr_nlwp;
		u = get_userstat(p.pr_uid);
		++u->nproc;
		u->nlwp += p.pr_nlwp;
		u->size += p.pr_size;
		u->rss += p.pr_rssize;
		u->pctcpu += (double)p.pr_pctcpu/0x8000*100;
		u->pctmem += (double)p.pr_pctmem/0x8000*100;
		u->time.tv_sec += p.pr_time.tv_sec;
		ns = u->time.tv_nsec + p.pr_time.tv_nsec;
		u->time.tv_sec += ns / 1000000000;
		u->time.tv_nsec = ns % 1000000000;
		assert((u->time.tv_sec >= 0) && (u->time.tv_nsec >= 0));
		switch(p.pr_lwp.pr_sname) {
		default:
			abort();
			break;
		case 'O':
			++u->oncpu;
			break;
		case 'S':
			++u->sleep;
			break;
		case 'R':
			++u->runnable;
			break;
		case 'Z':
			++u->zombie;
			break;
		case 'T':
			++u->stopped;
			break;
		}
		(void) close(psi);
	}
}


static int 
sort_pcpu(userstat_t *u1, userstat_t *u2)
{
	return u1->pctcpu < u2->pctcpu;
}


static int 
sort_nproc(userstat_t *u1, userstat_t *u2)
{
	return u1->nproc < u2->nproc;
}


static int 
sort_rss(userstat_t *u1, userstat_t *u2)
{
	return u1->rss < u2->rss;
}


static int 
sort_size(userstat_t *u1, userstat_t *u2)
{
	return u1->size < u2->size;
}


static int 
sort_time(userstat_t *u1, userstat_t *u2)
{
	return u1->time.tv_sec < u2->time.tv_sec;
}


static int 
sort_user(userstat_t *u1, userstat_t *u2)
{
	return strcmp(u1->username,u2->username) > 0;
}


void 
set_sort_by_pcpu(void)
{
	Crit = sort_pcpu;
}


void 
set_sort_by_nproc(void)
{
	Crit = sort_nproc;
}


void 
set_sort_by_rss(void)
{
	Crit = sort_rss;
}


void
set_sort_by_size(void)
{
	Crit = sort_size;
}


void
set_sort_by_time(void)
{
	Crit = sort_time;
}


void
set_sort_by_user(void)
{
	Crit = sort_user;
}


userstat_t *
sort_users(userstat_t *u)
{
	userstat_t *h = 0, *t = 0;
	while (u) {
		userstat_t *n = u->next, *i = h;
		if (u->nproc == 0) {
			free(u->username);
			free(u);
			u = n;
			continue;
		}
		while (i && Crit(u,i)
			&& i->next && Crit(u,i->next)) {
			assert(i != i->next);
			i = i->next;
		}
		if (i == 0) {
			if (t) {
				assert(h && !Crit(t, i));
				t->next = u;
				t = u;
			} else {
				assert(h == 0);
				h = t = u;
			}
			u->next = 0;
		} else {
			if (Crit(u,i)) {
				u->next = i->next;
				i->next = u;
			} else {
				u->next = i;
				h = u;
			}
			assert(i != i->next);
		}
		u = n;
		assert(u == 0 || u != u->next);
	}
	return h;
}


userstat_t *
get_procinfo(void)
{
	update_procinfo();
	UserHead = sort_users(UserHead);
	return UserHead;
}
